from functools import reduce # Valid in Python 2.6+, required in Python 3
import operator

import torch
import torch.nn as nn
import torch.nn.functional as F

import numpy as np
np.seterr(divide='ignore', invalid='ignore')

import torchvision
import torchvision.datasets as datasets

import os
import h5py

from mnist import fmnist,mnist,mnist_perf

import matplotlib.pyplot as plt

fmnist_L_x,fmnist_L_y = fmnist(True)
fmnist_T_x,fmnist_T_y = fmnist(False)

def lerp(a,b,λ): return λ*b + (1-λ)*a

device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')

class Pad(nn.Module):
	def __init__(self,*args):
		super().__init__()
		self.args=args
	def forward(self,x):
		return nn.functional.pad(x,*self.args)

def nn_fc(layers,σ,dropout=0.5):
	disable_drp = 1 if dropout==0 else 0
	return [
		nn.Linear(int(λ1),int(λ2)) if index==0 else
		σ                          if index==1 else
		nn.BatchNorm1d(int(λ2))    if index==2 else
		nn.Dropout(dropout)
		for λ1,λ2 in zip(layers[:-1],layers[1:])
		for index in range(4 - disable_drp)
	][:-3 + disable_drp] # remove last GELU, BatchNorm and Dropout

import random
class BatchIterator:
	def __init__(self,tensors,batch_size,shuffle=True):
		self.tensors = tensors # must be a tuple of tensor
		self.batch_size = batch_size
		# sanity check
		if not len(set([t.shape[0] for t in tensors])) == 1: print("ERROR : inconsistent input")
		self.indices = list(range(tensors[0].shape[0]))
		if shuffle: random.shuffle(self.indices)
	def __iter__(self): return self
	def __next__(self):
		batch_size= min(self.batch_size,len(self.indices))
		if batch_size==0: raise StopIteration
		indices = [self.indices.pop(0) for s in range(batch_size)]
		return [torch.stack([t[i] for i in indices],0) for t in self.tensors]

# xs = torch.randn(10,1,28,28)
# ys = torch.randn(10,10)
# for i,(x,y) in enumerate(BatchIterator((xs,ys),3)):
# 	print(x.shape,y.shape)

# def analysis_classification(model,identifier,lr0=-2,lr1=-4,epochs=256,batch_size=1024):
def analysis_classification(model,identifier,loss_fun=nn.MSELoss(),lr0=-2,lr1=-4,epochs=128,batch_size=64):
	model.to(device)
	model.identifier=identifier
	optim = torch.optim.Adam(model.parameters(),lr=10**lr0)
	hist = {
		"epoch_L":[], "loss_L": [], "loss_L_μ": [], "loss_L_σ": [], "perf_L_precision": [], "perf_L_accuracy": [], "perf_L_recall": [], "perf_L_F1": [],
		"epoch_T":[], "loss_T": [], "loss_T_μ": [], "loss_T_σ": [], "perf_T_precision": [], "perf_T_accuracy": [], "perf_T_recall": [], "perf_T_F1": [],
	}

	dataset_L = torch.utils.data.DataLoader(list(zip(fmnist_L_x,fmnist_L_y)), batch_size=batch_size, shuffle=True)
	dataset_T = torch.utils.data.DataLoader(list(zip(fmnist_T_x,fmnist_T_y)), batch_size=batch_size, shuffle=True)

	for epoch in range(epochs):
		lr = 10**lerp(lr0,lr1,epoch/(epochs-1))
		for g in optim.param_groups: g['lr'] = lr
		#
		model.train()
		losses=[]
		# for i,(x,y) in enumerate(BatchIterator((fmnist_L_x,fmnist_L_y),batch_size)):
		for i,(x,y) in enumerate(dataset_L):
			x=x.to(device)
			y=y.to(device)
			loss = loss_fun(model(x),y)
			optim.zero_grad();loss.requires_grad_().backward();optim.step()
			losses.append(loss.item())
			hist["loss_L"].append(loss.item())
			hist["epoch_L"].append(epoch + i*batch_size/fmnist_L_x.shape[0])
		hist["loss_L_μ"].append(np.mean(losses))
		hist["loss_L_σ"].append(np.std(losses))
		#
		model.eval()
		losses=[]
		# for i,(x,y) in enumerate(BatchIterator((fmnist_T_x,fmnist_T_y),batch_size)):
		for i,(x,y) in enumerate(dataset_T):
			x=x.to(device)
			y=y.to(device)
			loss = loss_fun(model(x),y)
			losses.append(loss.item())
			hist["loss_T"].append(loss.item())
			hist["epoch_T"].append(epoch + i*batch_size/fmnist_T_x.shape[0])
		hist["loss_T_μ"].append(np.mean(losses))
		hist["loss_T_σ"].append(np.std(losses))
		#
		rgb_L = (0,.5,1)
		rgb_T = (1,.5,0)
		#
		pr_L,re_L,ac_L,F1_L,cm_L=mnist_perf(model,device,"learn",fmnist_L_x,fmnist_L_y,batch_size,rgb_L) ### on learn set
		hist["perf_L_precision"].append(pr_L)
		hist["perf_L_recall"   ].append(re_L)
		hist["perf_L_accuracy" ].append(ac_L)
		hist["perf_L_F1"       ].append(F1_L)
		#
		pr_T,re_T,ac_T,F1_T,cm_T=mnist_perf(model,device,"test" ,fmnist_T_x,fmnist_T_y,batch_size,rgb_T) ### on test set
		hist["perf_T_precision"].append(pr_T)
		hist["perf_T_recall"   ].append(re_T)
		hist["perf_T_accuracy" ].append(ac_T)
		hist["perf_T_F1"       ].append(F1_T)
		os.makedirs(f"out/",exist_ok=True)
		print(f"epoch={epoch+1:3d}/{epochs:3d} | log(loss)={np.log(hist['loss_L_μ'][-1]):.5f} | L_prec={pr_L:.5f} L_recall={re_L:.5f} L_accur={ac_L:.5f} L_F1={F1_L:.5f} | T_prec={pr_T:.5f} T_recall={re_T:.5f} T_accur={ac_T:.5f} T_F1={F1_T:.5f}")
		#
		plt.figure()
		plt.scatter(hist["epoch_L"],np.log(hist['loss_L']), label='log(loss) train', s=0.05, alpha=1/4, color=rgb_L)
		plt.scatter(hist["epoch_T"],np.log(hist['loss_T']), label='log(loss) test' , s=0.05, alpha=1/4, color=rgb_T)
		plt.legend()
		plt.tight_layout();plt.savefig(f"out/{model.identifier}-loss.png");plt.close('all')
		#
		plt.figure()
		plt.plot(hist['perf_L_accuracy' ], label=f'learn accuracy' , color=rgb_L , linestyle='-')
		plt.plot(hist['perf_L_precision'], label=f'learn precision', color=rgb_L, linestyle='--')
		plt.plot(hist['perf_L_recall'   ], label=f'learn recall'   , color=rgb_L, linestyle=':') # TODO : why perf_L_recall == perf_L_accuracy ?
		plt.plot(hist['perf_L_F1'       ], label=f'learn F1'       , color=rgb_L, linestyle='-.')
		plt.plot(hist['perf_T_accuracy' ], label=f'test accuracy'  , color=rgb_T, linestyle='-')
		plt.plot(hist['perf_T_precision'], label=f'test precision' , color=rgb_T, linestyle='--')
		plt.plot(hist['perf_T_recall'   ], label=f'test recall'    , color=rgb_T, linestyle=':') # TODO : why perf_T_recall == perf_T_accuracy ?
		plt.plot(hist['perf_T_F1'       ], label=f'test F1'        , color=rgb_T, linestyle='-.')
		plt.legend()
		plt.tight_layout();plt.savefig(f"out/{model.identifier}-perf.png");plt.close('all')
		#
		plt.figure()
		plt.plot(range(epoch+1),hist['loss_L_μ'], label='learn loss', color=rgb_L)
		plt.plot(range(epoch+1),hist['loss_T_μ'], label='test loss' , color=rgb_T)
		plt.fill_between(range(epoch+1), np.array(hist['loss_L_μ'])-np.array(hist['loss_L_σ']), np.array(hist['loss_L_μ'])+np.array(hist['loss_L_σ']), alpha=0.2, color=rgb_L)
		plt.fill_between(range(epoch+1), np.array(hist['loss_T_μ'])-np.array(hist['loss_T_σ']), np.array(hist['loss_T_μ'])+np.array(hist['loss_T_σ']), alpha=0.2, color=rgb_T)
		plt.legend()
		plt.yscale('log')
		plt.tight_layout();plt.savefig(f"out/{model.identifier}-loss-2.png");plt.close('all')
		#
		with h5py.File(f'out/{model.identifier}.h5', 'w') as file:
			for k in hist: file.create_dataset(k, data=hist[k])

def analysis_auto_enc(model,identifier,lr0,lr1,epochs=64,batch_size=64):
	pass # TODO

# for loss_name,loss in [
# 		["MSE",nn.MSELoss()],
# 	]:
# 	for σ_name,σ in σs:
# 		for N in [4]:
# 			for m_size in [2]:
# 				for M in [1,2,4,8,16]:
# 					for ker_size in [3]:
# 						for dropout in [0,0.25,0.5,0.75]:
# 							add_job(lambda: analysis_classification(
# 								nn_cn(N,m_size,M,ker_size),
# 								f"cn({N},{m_size},{M},{ker_size})-{σ_name}-dropout={dropout}-loss={loss_name}",
# 								loss
# 							))
